#include <sstream>
#include <algorithm>
#include <stdexcept>

#include "npstat/stat/AbsSmoothGOFTest1D.hh"
#include "npstat/stat/Distributions1D.hh"

namespace npstat {
    AbsSmoothGOFTest1D::AbsSmoothGOFTest1D(const AbsDistribution1D& nullDistro,
                                           const unsigned maxdeg,
                                           const unsigned char* mask,
                                           const unsigned lenMask)
        : AbsUnbinnedGOFTest1D(nullDistro),
          mask_(maxdeg, 0),
          maxdeg_(maxdeg),
          nUnmasked_(maxdeg),
          devBuf_(maxdeg)
    {
        initMask(mask, lenMask);
    }

    AbsSmoothGOFTest1D::AbsSmoothGOFTest1D(const AbsDistribution1D& nullDistro,
                                           const std::vector<int>& mask)
        : AbsUnbinnedGOFTest1D(nullDistro),
          mask_(mask.size(), 0),
          maxdeg_(mask.size()),
          nUnmasked_(maxdeg_),
          devBuf_(maxdeg_)
    {
        if (!mask.empty())
            initMask(&mask[0], mask.size());
        else
            initMask((int*)0, 0);
    }

    double AbsSmoothGOFTest1D::testStatistic(
        const double* data, const unsigned long sz,
        const bool isDataSorted) const
    {
        const unsigned len = this->normalizedDeviations(
            data, sz, isDataSorted,
            &devBuf_[0], devBuf_.size());
        return sumOfSquares(&devBuf_[0], len);
    }

    double AbsSmoothGOFTest1D::testStatistic(
        const float* data, const unsigned long sz,
        const bool isDataSorted) const
    {
        const unsigned len = this->normalizedDeviations(
            data, sz, isDataSorted,
            &devBuf_[0], devBuf_.size());
        return sumOfSquares(&devBuf_[0], len);
    }

    double AbsSmoothGOFTest1D::analyticPValue(
        const double stat, unsigned long /* sz */) const
    {
        Gamma1D chisq(0.0, 2.0, nUnmasked_/2.0);
        return chisq.exceedance(stat);
    }

    double AbsSmoothGOFTest1D::inverseExceedance(
        const double pvalue, const unsigned long sz,
        const double smin, const double smax) const
    {
        if (pvalue <= 0.0 || pvalue >= 1.0) throw std::invalid_argument(
            "In npstat::AbsSmoothGOFTest1D::inverseExceedance: "
            "p-value argument must be inside (0, 1) interval");

        if (pvalue > 1.0e-10)
        {
            Gamma1D chisq(0.0, 2.0, nUnmasked_/2.0);
            return chisq.quantile(1.0 - pvalue);
        }
        else
            return AbsUnbinnedGOFTest1D::inverseExceedance(
                pvalue, sz, smin, smax);
    }

    std::string AbsSmoothGOFTest1D::testNameWithMask(const char* prefix) const
    {
        std::ostringstream os;
        os << prefix << '_' << maxdeg_;
        if (nUnmasked_ < maxdeg_)
        {
            // Figure out the number of the last
            // masked term
            unsigned lastMasked = maxdeg_;
            for (unsigned i=0; i<maxdeg_; ++i)
                if (mask_[i])
                    lastMasked = i;
            assert(lastMasked < maxdeg_);

            // Add mask info
            os << '_';
            for (unsigned i=0; i<=lastMasked; ++i)
            {
                if (mask_[i])
                    os << '1';
                else
                    os << '0';
            }            
        }
        return os.str();
    }

    void AbsSmoothGOFTest1D::simulateAltDirectedStatistic(
        const double* direction, const unsigned lenDirection,
        const AbsDistribution1D& alternative,
        AbsRandomGenerator& g, const unsigned long sz,
        const unsigned nPseudo, std::vector<double>* stats) const
    {
        if (!sz) throw std::invalid_argument(
            "In npstat::AbsSmoothGOFTest1D::simulateAltStatistic: "
            "sample size must be positive");
        if (!nPseudo) std::invalid_argument(
            "In npstat::AbsSmoothGOFTest1D::simulateAltStatistic: "
            "number of pseudo experiments must be positive");
        assert(stats);
        stats->clear();
        stats->reserve(nPseudo);

        std::vector<double> sampleVec(sz);
        double* sample = &sampleVec[0];
        for (unsigned ips=0; ips<nPseudo; ++ips)
        {
            for (unsigned long i=0; i<sz; ++i)
                alternative.random(g, sample+i);
            std::sort(sampleVec.begin(), sampleVec.end());
            stats->push_back(this->directedStatistic(
                direction, lenDirection, sample, sz, true));
        }
        std::sort(stats->begin(), stats->end());
    }
}
